import numpy as np

from menpo.transform import AlignmentAffine, AlignmentSimilarity, Rotation
from menpo.shape import PointCloud

import menpo3d.checks as checks
from menpo3d.camera import PerspectiveCamera, OrthographicCamera

from .algorithm import SimultaneousForwardAdditive
from .result import MMResult
from .shapemodel import ShapeModel


class MMFitter(object):
    r"""
    Abstract class for defining a multi-scale Morphable Model fitter.

    Parameters
    ----------
    mm : :map:`ColouredMorphableModel` or :map:`TexturedMorphableModel`
        The trained Morphable Model.
    algorithms : `list` of `class`
        The list of algorithm objects that will perform the fitting per scale.
    camera_cls : `menpo3d.camera.PerspectiveCamera` or `menpo3d.camera.OrthographicCamera`
        The camera class to use.
    """

    def __init__(self, mm, algorithms, camera_cls):
        # Assign model and algorithms
        self._model = mm
        self.algorithms = algorithms
        self.n_scales = len(self.algorithms)
        self.camera_cls = camera_cls

    @property
    def mm(self):
        r"""
        The trained Morphable Model.

        :type: :map:`ColouredMorphableModel` or `:map:`TexturedMorphableModel`
        """
        return self._model

    @property
    def holistic_features(self):
        r"""
        The features that are extracted from the input image.

        :type: `function`
        """
        return self.mm.holistic_features

    @property
    def diagonal(self):
        r"""
        The diagonal used to rescale the image.

        :type: `int`
        """
        return self.mm.diagonal

    def _prepare_image(self, image, initial_shape):
        r"""
        Function the performs pre-processing on the image to be fitted. This
        involves the following steps:

            1. Rescale image so that the provided initial_shape has the
               specified diagonal.
            2. Compute features
            3. Estimate the affine transform introduced by the rescale to
               diagonal and features extraction

        Parameters
        ----------
        image : `menpo.image.Image` or subclass
            The image to be fitted.
        initial_shape : `menpo.shape.PointCloud`
            The initial shape estimate from which the fitting procedure
            will start.

        Returns
        -------
        image : `menpo.image.Image`
            The feature-based image.
        initial_shape : `menpo.shape.PointCloud`
            The rescaled initial shape.
        affine_transform : `menpo.transform.Affine`
            The affine transform that is the inverse of the transformations
            introduced by the rescale wrt diagonal as well as the feature
            extraction.
        """
        # Attach landmarks to the image, in order to make transforms easier
        image.landmarks["__initial_shape"] = initial_shape

        if self.diagonal is not None:
            # Rescale image so that initial_shape matches the provided diagonal
            tmp_image = image.rescale_landmarks_to_diagonal_range(
                self.diagonal, group="__initial_shape"
            )
        else:
            tmp_image = image
        # Extract features
        feature_image = self.holistic_features(tmp_image)

        # Get final transformed landmarks
        new_initial_shape = feature_image.landmarks["__initial_shape"]

        # Now we have introduced an affine transform that consists of the image
        # rescaled based on the diagonal, as well as potential rescale
        # (down-sampling) caused by features. We need to store this transform
        # (estimated by AlignmentAffine) in order to be able to revert it at
        # the final fitting result.
        affine_transform = AlignmentAffine(new_initial_shape, initial_shape)

        # Detach added landmarks from image
        del image.landmarks["__initial_shape"]

        return feature_image, new_initial_shape, affine_transform

    def _align_mean_shape_with_bbox(self, bbox):
        # Convert 3D landmarks to 2D by removing the Z axis
        template_shape = PointCloud(self.mm.landmarks.points[:, [1, 0]])

        # Rotation that flips over x axis
        rot_matrix = np.eye(template_shape.n_dims)
        rot_matrix[0, 0] = -1
        template_shape = Rotation(rot_matrix, skip_checks=True).apply(template_shape)

        # Align the 2D landmarks' bbox with the provided bbox
        return AlignmentSimilarity(template_shape.bounding_box(), bbox).apply(
            template_shape
        )

    def _fit(
        self,
        image,
        camera,
        instance=None,
        gt_mesh=None,
        max_iters=50,
        camera_update=False,
        focal_length_update=False,
        reconstruction_weight=1.0,
        shape_prior_weight=None,
        texture_prior_weight=None,
        landmarks_prior_weight=None,
        landmarks=None,
        return_costs=False,
    ):
        # Check provided instance
        if instance is None:
            instance = self.mm.instance()

        # Check arguments
        max_iters = checks.check_max_iters(max_iters, self.n_scales)
        reconstruction_weight = checks.check_multi_scale_param(
            self.n_scales,
            (float, int, None),
            "reconstruction_prior_weight",
            reconstruction_weight,
        )
        shape_prior_weight = checks.check_multi_scale_param(
            self.n_scales, (float, int, None), "shape_prior_weight", shape_prior_weight
        )
        texture_prior_weight = checks.check_multi_scale_param(
            self.n_scales,
            (float, int, None),
            "texture_prior_weight",
            texture_prior_weight,
        )
        landmarks_prior_weight = checks.check_multi_scale_param(
            self.n_scales,
            (float, int, None),
            "landmarks_prior_weight",
            landmarks_prior_weight,
        )
        for i in range(self.n_scales):
            if reconstruction_weight[i] is None and landmarks_prior_weight[i] is None:
                reconstruction_weight[i] = 1.0
            if reconstruction_weight[i] is None:
                texture_prior_weight[i] = None

        # Initialize list of algorithm results
        algorithm_results = []

        # Main loop at each scale level
        for i in range(self.n_scales):
            # Run algorithm
            algorithm_result = self.algorithms[i].run(
                image,
                instance,
                camera,
                gt_mesh=gt_mesh,
                max_iters=max_iters[i],
                camera_update=camera_update,
                focal_length_update=focal_length_update,
                reconstruction_weight=reconstruction_weight[i],
                shape_prior_weight=shape_prior_weight[i],
                texture_prior_weight=texture_prior_weight[i],
                landmarks_prior_weight=landmarks_prior_weight[i],
                landmarks=landmarks,
                return_costs=return_costs,
            )

            # Get current instance
            instance = algorithm_result.final_mesh
            camera = algorithm_result.final_camera_transform

            # Add algorithm result to the list
            algorithm_results.append(algorithm_result)

        return algorithm_results

    def fit_from_camera(
        self,
        image,
        camera,
        instance=None,
        gt_mesh=None,
        max_iters=50,
        camera_update=False,
        focal_length_update=False,
        shape_prior_weight=1.0,
        texture_prior_weight=1.0,
        return_costs=False,
    ):
        # Execute multi-scale fitting
        algorithm_results = self._fit(
            image,
            camera,
            instance=instance,
            gt_mesh=gt_mesh,
            max_iters=max_iters,
            camera_update=camera_update,
            focal_length_update=focal_length_update,
            reconstruction_weight=1.0,
            shape_prior_weight=shape_prior_weight,
            texture_prior_weight=texture_prior_weight,
            landmarks_prior_weight=None,
            landmarks=None,
            return_costs=return_costs,
        )

        # Return multi-scale fitting result
        return self._fitter_result(
            image=image,
            algorithm_results=algorithm_results,
            affine_transform=Rotation.init_identity(n_dims=2),
            gt_mesh=gt_mesh,
        )

    def fit_from_bb(
        self,
        image,
        initial_bb,
        gt_mesh=None,
        max_iters=50,
        camera_update=False,
        focal_length_update=False,
        reconstruction_weight=1.0,
        shape_prior_weight=1.0,
        texture_prior_weight=1.0,
        return_costs=False,
        distortion_coeffs=None,
    ):
        initial_shape = self._align_mean_shape_with_bbox(initial_bb)
        return self.fit_from_shape(
            image=image,
            initial_shape=initial_shape,
            gt_mesh=gt_mesh,
            max_iters=max_iters,
            camera_update=camera_update,
            focal_length_update=focal_length_update,
            reconstruction_weight=reconstruction_weight,
            shape_prior_weight=shape_prior_weight,
            texture_prior_weight=texture_prior_weight,
            landmarks_prior_weight=None,
            return_costs=return_costs,
            distortion_coeffs=distortion_coeffs,
            init_shape_params_from_lms=False,
        )

    def fit_from_shape(
        self,
        image,
        initial_shape,
        gt_mesh=None,
        max_iters=50,
        camera_update=True,
        focal_length_update=False,
        reconstruction_weight=1.0,
        shape_prior_weight=1.0,
        texture_prior_weight=1.0,
        landmarks_prior_weight=1.0,
        return_costs=False,
        distortion_coeffs=None,
        init_shape_params_from_lms=False,
    ):
        # Check that the provided initial shape has the same number of points
        # as the landmarks of the model
        if initial_shape.n_points != self.mm.landmarks.n_points:
            raise ValueError(
                "The provided 2D initial shape must have {} landmark "
                "points.".format(self.mm.landmarks.n_points)
            )

        # Rescale image and extract features
        rescaled_image, rescaled_initial_shape, affine_transform = self._prepare_image(
            image, initial_shape
        )

        # Estimate view, projection and rotation transforms from the
        # provided initial shape
        camera = self.camera_cls.init_from_2d_projected_shape(
            self.mm.landmarks,
            rescaled_initial_shape,
            rescaled_image.shape,
            distortion_coeffs=distortion_coeffs,
        )

        if init_shape_params_from_lms:
            # Wrap the shape model in a container that allows us to mask the
            # PCA basis spatially
            sm = ShapeModel(self.mm.shape_model)
            # Only keep the landmark points in the basis
            sm_lms_3d = sm.mask_points(self.mm.model_landmarks_index)

            # Warp the basis with the camera and retain only the first two dims
            sm_lms_2d = camera.apply(sm_lms_3d).mask_dims([0, 1])
            # Project onto the first few shape components to give an initial
            # shape
            shape_weights = sm_lms_2d.project(rescaled_initial_shape, n_components=10)
            instance = self.mm.instance(shape_weights)
        else:
            instance = None

        # Execute multi-scale fitting
        algorithm_results = self._fit(
            rescaled_image,
            camera,
            instance=instance,
            gt_mesh=gt_mesh,
            max_iters=max_iters,
            camera_update=camera_update,
            focal_length_update=focal_length_update,
            reconstruction_weight=reconstruction_weight,
            shape_prior_weight=shape_prior_weight,
            texture_prior_weight=texture_prior_weight,
            landmarks_prior_weight=landmarks_prior_weight,
            landmarks=rescaled_initial_shape,
            return_costs=return_costs,
        )

        # Return multi-scale fitting result
        return self._fitter_result(
            image=image,
            algorithm_results=algorithm_results,
            affine_transform=affine_transform,
            gt_mesh=gt_mesh,
        )

    def _fitter_result(self, image, algorithm_results, affine_transform, gt_mesh=None):
        return MMResult(
            algorithm_results,
            [affine_transform] * self.n_scales,
            self.n_scales,
            image=image,
            gt_mesh=gt_mesh,
            model_landmarks_index=self.mm.model_landmarks_index,
        )


class LucasKanadeMMFitter(MMFitter):
    def __init__(
        self,
        mm,
        lk_algorithm_cls=SimultaneousForwardAdditive,
        n_scales=1,
        n_shape=1.0,
        n_texture=1.0,
        n_samples=1000,
        camera_cls=PerspectiveCamera,
    ):
        # Check parameters
        n_shape = checks.check_multi_scale_param(n_scales, (int,), "n_shape", n_shape)
        n_texture = checks.check_multi_scale_param(
            n_scales, (int,), "n_texture", n_texture
        )
        self.n_samples = checks.check_multi_scale_param(
            n_scales, (int,), "n_samples", n_samples
        )
        if camera_cls in [PerspectiveCamera, OrthographicCamera]:
            self.camera_cls = camera_cls
        else:
            raise ValueError(
                "Camera can be either PerspectiveCamera or" "OrthographicCamera"
            )

        # Create list of algorithms
        algorithms = []
        for i in range(n_scales):
            mm_copy = mm.copy()
            set_model_components(mm_copy.shape_model, n_shape[i])
            set_model_components(mm_copy.texture_model, n_texture[i])
            algorithms.append(lk_algorithm_cls(mm_copy, self.n_samples[i]))

        # Call superclass
        super(LucasKanadeMMFitter, self).__init__(
            mm=mm, algorithms=algorithms, camera_cls=camera_cls
        )

    def __str__(self):
        scales_info = []
        lvl_str_tmplt = r"""   - Scale {}
     - {} active shape components
     - {} active texture components"""
        for k in range(self.n_scales):
            scales_info.append(
                lvl_str_tmplt.format(
                    k,
                    self.algorithms[k].model.shape_model.n_active_components,
                    self.algorithms[k].model.texture_model.n_active_components,
                )
            )
        scales_info = "\n".join(scales_info)

        cls_str = r"""{class_title}
 - {scales} scales
{scales_info}
    """.format(
            class_title=self.algorithms[0].__str__(),
            scales=self.n_scales,
            scales_info=scales_info,
        )
        return self.mm.__str__() + cls_str


def set_model_components(model, n_components):
    r"""
    Function that sets the number of active components to the provided model.

    Parameters
    ----------
    model : `menpo.model.PCAVectorModel` or `menpo.model.PCAModel` or subclass
        The PCA model.
    n_components : `int` or `float` or ``None``
        The number of active components to set.

    Raises
    ------
    ValueError
        n_components can be an integer or a float or None
    """
    if n_components is not None:
        if type(n_components) is int or type(n_components) is float:
            model.n_active_components = n_components
        else:
            raise ValueError("n_components can be an integer or a float or " "None")



import numpy as np
from collections import namedtuple

optimise = None  # expensive, from scipy

RadialFitResult = namedtuple("RadialFitResult", ["centre", "radius"])


def radial_fit(p):
    """
    Find the least squares radial fitting a set of ND points.

    Parameters
    ----------
    p : ``(N, D)`` `ndarray`
        Points to find find centre of

    Returns
    -------
    centre_i : (D,) `ndarray`
        The ND coordinates of the centre of the circle.
    r_i : `float`
        The radius of the circle.

    References
    ----------
    .. [1] http://www.scipy.org/Cookbook/Least_Squares_Circle

    """
    global optimise
    if optimise is None:
        from scipy import optimize  # expensive

    def error(tuple_c, x):
        c = np.array(tuple_c)
        err = r(x, c)
        return err - err.mean()

    r = lambda x, c: np.sqrt(np.sum((x - c) ** 2, axis=1))
    av_r = lambda x, c: np.mean(r(x, c))
    c_est = np.mean(p, axis=0)
    c_i, ier = optimize.leastsq(error, c_est, args=p)
    return RadialFitResult(centre=c_i, radius=av_r(p, c_i))



import os
import platform
import site

from setuptools import Extension, find_packages, setup

SYS_PLATFORM = platform.system().lower()
IS_LINUX = "linux" in SYS_PLATFORM
IS_OSX = "darwin" == SYS_PLATFORM
IS_WIN = "windows" == SYS_PLATFORM

# Get Numpy include path without importing it
NUMPY_INC_PATHS = [
    os.path.join(r, "numpy", "core", "include")
    for r in site.getsitepackages()
    if os.path.isdir(os.path.join(r, "numpy", "core", "include"))
]
if len(NUMPY_INC_PATHS) == 0:
    try:
        import numpy as np
    except ImportError:
        raise ValueError(
            "Could not find numpy include dir and numpy not installed before build - "
            "cannot proceed with compilation of cython modules."
        )
    else:
        # just ask numpy for it's include dir
        NUMPY_INC_PATHS = [np.get_include()]

elif len(NUMPY_INC_PATHS) > 1:
    print(
        "Found {} numpy include dirs: "
        "{}".format(len(NUMPY_INC_PATHS), ", ".join(NUMPY_INC_PATHS))
    )
    print("Taking first (highest precedence on path): {}".format(NUMPY_INC_PATHS[0]))
NUMPY_INC_PATH = NUMPY_INC_PATHS[0]


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location("version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


# ---- C/C++ EXTENSIONS ---- #
# Stolen (and modified) from the Cython documentation:
#     http://cython.readthedocs.io/en/latest/src/reference/compilation.html
def no_cythonize(extensions, **_ignore):
    import os.path as op

    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
                if not op.exists(sfile):
                    raise ValueError(
                        "Cannot find pre-compiled source file "
                        "({}) - please install Cython".format(sfile)
                    )
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


def build_extension_from_pyx(pyx_path, extra_sources_paths=None):
    if extra_sources_paths is None:
        extra_sources_paths = []
    extra_sources_paths.insert(0, pyx_path)
    ext = Extension(
        name=pyx_path[:-4].replace("/", "."),
        sources=extra_sources_paths,
        include_dirs=[NUMPY_INC_PATH],
        language="c++",
    )
    if IS_LINUX or IS_OSX:
        ext.extra_compile_args.append("-Wno-unused-function")
    if IS_OSX:
        ext.extra_link_args.append("-headerpad_max_install_names")
    return ext


try:
    from Cython.Build import cythonize
except ImportError:
    import warnings

    cythonize = no_cythonize
    warnings.warn(
        "Unable to import Cython - attempting to build using the "
        "pre-compiled C++ files."
    )

cython_modules = [
    build_extension_from_pyx("menpo3d/rasterize/tripixel.pyx"),
]
cython_exts = cythonize(cython_modules, quiet=True)

version, cmdclass = get_version_and_cmdclass("menpo3d")

install_requires = ["menpo>=0.9.0,<0.12.0", "mayavi>=4.7.0", "moderngl>=5.6.*,<6.0"]

setup(
    name="menpo3d",
    version=version,
    cmdclass=cmdclass,
    description="Menpo library providing tools for 3D Computer Vision research",
    author="James Booth",
    author_email="james.booth08@imperial.ac.uk",
    packages=find_packages(),
    package_data={
        "menpo3d": [
            "data/*",
            "rasterize/shaders/per_vertex/*",
            "rasterize/shaders/texture/*",
        ]
    },
    install_requires=install_requires,
    tests_require=["pytest>=5.0"],
    ext_modules=cython_exts,
)



import menpo.io as mio
import menpo3d.io as m3dio
import numpy as np
from menpo3d.correspond import solve_pnp_landmarks


def test_solvepnp():
    template = m3dio.import_builtin_asset.template_ply()
    image = mio.import_builtin_asset.lenna_png()

    # Comes from running 100 times and taking the mean
    expected = np.array(
        [
            [0.9188, 0.1053, 0.3804, -622.7377],
            [0.0395, -0.9834, 0.1769, 28059.479],
            [0.3927, -0.1475, -0.9077, 512972.2382],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    # Ransac is random and we can't control the seed so we do a fuzzy test
    model_view_t, _ = solve_pnp_landmarks(image, template, group="LJSON")
    np.testing.assert_allclose(model_view_t.h_matrix, expected, atol=1e-2)



from scipy.io import loadmat
from menpo.model import PCAModel
from menpo.shape import TriMesh


def lsfm_model_importer(path, **kwargs):
    m = loadmat(str(path))
    mean = TriMesh(m["mean"].reshape([-1, 3]), trilist=m["trilist"])
    return PCAModel.init_from_components(
        m["components"].T, m["eigenvalues"].ravel(), mean, m["n_training_samples"], True
    )



import collections
import numpy as np


def check_max_iters(max_iters, n_scales):
    r"""
    Function that checks the value of a `max_iters` parameter defined for
    multiple scales. It must be `int` or `list` of `int`.

    Parameters
    ----------
    max_iters : `int` or `list` of `int`
        The value to check.
    n_scales : `int`
        The number of scales.

    Returns
    -------
    max_iters : `list` of `int`
        The list of values per scale.

    Raises
    ------
    ValueError
        max_iters can be integer, integer list containing 1 or {n_scales}
        elements or None
    """
    if type(max_iters) is int:
        max_iters = [np.round(max_iters / n_scales) for _ in range(n_scales)]
    elif len(max_iters) == 1 and n_scales > 1:
        max_iters = [np.round(max_iters[0] / n_scales) for _ in range(n_scales)]
    elif len(max_iters) != n_scales:
        raise ValueError(
            "max_iters can be integer, integer list "
            "containing 1 or {} elements or "
            "None".format(n_scales)
        )
    return np.require(max_iters, dtype=np.int)


def check_multi_scale_param(n_scales, types, param_name, param):
    r"""
    General function for checking a parameter defined for multiple scales. It
    raises an error if the parameter is not an iterable with the correct size and
    correct types.

    Parameters
    ----------
    n_scales : `int`
        The number of scales.
    types : `tuple`
        The `tuple` of variable types that the parameter is allowed to have.
        It can also contain ``None``.
    param_name : `str`
        The name of the parameter.
    param : `types` or `list` of `types`
        The parameter value.

    Returns
    -------
    param : `list` of `types`
        The list of values per scale.

    Raises
    ------
    ValueError
        {param_name} must be in {types} or a list/tuple of {types} with the same
        length as the number of scales
    """
    error_msg = (
        "{0} must be in {1} or a list/tuple of "
        "{1} with the same length as the number "
        "of scales".format(param_name, types)
    )

    none_in_types = False
    new_types = []
    for i in types:
        if i is None:
            none_in_types = True
        else:
            new_types.append(i)
    new_types = tuple(new_types)

    # Could be a single value - or we have an error
    if none_in_types:
        if isinstance(param, new_types) or param is None:
            return [param] * n_scales
        elif not isinstance(param, collections.Iterable):
            raise ValueError(error_msg)
    else:
        if isinstance(param, new_types):
            return [param] * n_scales
        elif not isinstance(param, collections.Iterable):
            raise ValueError(error_msg)

    # Must be an iterable object
    len_param = len(param)
    if none_in_types:
        isinstance_all_in_param = all(
            isinstance(p, new_types) or p is None for p in param
        )
    else:
        isinstance_all_in_param = all(isinstance(p, new_types) for p in param)

    if len_param == 1 and isinstance_all_in_param:
        return list(param) * n_scales
    elif len_param == n_scales and isinstance_all_in_param:
        return list(param)
    else:
        raise ValueError(error_msg)


def check_parameters(parameters, n_active_components):
    params = parameters
    if params is None:
        params = np.zeros(n_active_components)
    elif len(params) > n_active_components:
        params = np.asarray(params[:n_active_components])
    elif len(params) < n_active_components:
        tmp = params
        params = np.zeros(n_active_components)
        params[: len(tmp)] = np.asarray(tmp)
    return params
